package de.fhdw.gaming.ipspiel23.ht.domain.impl;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import de.fhdw.gaming.core.domain.AbstractPlayer;
import de.fhdw.gaming.ipspiel23.ht.domain.Answer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTPlayer;

/**
 * Represents an Heads or Tails player implementation.
 */
final class HTPlayer extends AbstractPlayer<IHTPlayer> implements IHTPlayer {

    /**
     * The possible outcomes of the player.
     */
    private final Map<Answer, Map<Answer, Double>> possibleOutcomes;

    /**
     * The answer of the player.
     */
    private Optional<Answer> answer = Optional.empty();

    /**
     * Creates a new instance of {@link HTPlayer}.
     *
     * @param name           The name of the player.
     * @param possibleOutcomes The possible outcomes of the player.
     */
    protected HTPlayer(final String name, final Map<Answer, Map<Answer, Double>> possibleOutcomes) {
        super(name);
        this.possibleOutcomes = Collections
            .unmodifiableMap(new LinkedHashMap<>(Objects
                .requireNonNull(possibleOutcomes, "possibleOutcomes")));
    }

    /**
     * Creates a new instance of {@link HTPlayer}.
     *
     * @param source The source player.
     */
    protected HTPlayer(final IHTPlayer source) {
        super(source);
        possibleOutcomes = source.getPossibleOutcomes();
        answer = source.getAnswer();
    }

    @Override
    public IHTPlayer deepCopy() {
        return new HTPlayer(this);
    }

    @Override
    public Map<Answer, Map<Answer, Double>> getPossibleOutcomes() {
        return possibleOutcomes;
    }

    @Override
    public Optional<Answer> getAnswer() {
        return answer;
    }

    @Override
    public void setAnswer(final Answer newAnswer) {
        if (answer.isEmpty()) {
            answer = Optional.of(newAnswer);
            return;
        }
        throw new IllegalStateException(String.format("Player '%s' tried to change their answer.", this.getName()));
    }

    @Override
    public String toString() {
        return String
            .format("%s[name=%s, state=%s, outcome=%s, answer=%s]", 
                this.getClass().getSimpleName(), this.getName(), this.getState(),
                this.getOutcome(),
                this.answer);
    }
    
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof HTPlayer) {
            final HTPlayer other = (HTPlayer) obj;
            return super.equals(obj) && this.answer.equals(other.answer)
                && this.possibleOutcomes.equals(other.possibleOutcomes);
        }
        return false;
    }

    @Override
    public int hashCode() {
        return super.hashCode() ^ Objects.hash(this.answer, this.possibleOutcomes);
    }
}