package de.fhdw.gaming.ipspiel23.ht.domain.impl;

import java.util.Collections;
import java.util.Map;
import java.util.Optional;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.ipspiel23.ht.domain.Answer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTPlayer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTPlayerBuilder;

/**
 * A builder which implements the {@link IHTPlayerBuilder} and allows to create an {@link IHTPlayer}.
 */
final class HTPlayerBuilder implements IHTPlayerBuilder {

    /**
     * The name of the player.
     */
    private Optional<String> name;

    /**
     * The possible outcomes of the player.
     */
    private Optional<Map<Answer, Map<Answer, Double>>> possibleOutcomes;

    /**
     * Change the name of the player.
     * <p>
     * There is no default.
     * 
     * @param newName The new name of the player.
     * @return {@code this}
     */
    @Override
    public IHTPlayerBuilder changeName(final String newName) {
        name = Optional.of(newName);
        return this;
    }

    /**
     * Changes the possible outcomes of the player.
     * <p>
     * There is no default.
     *
     * @param newOutcomes The possible outcomes of the player. The key for the first-level map is the answer of the
     *                         first player, the key for the second-level map is the answer of the second player.
     */
    @Override
    public IHTPlayerBuilder changePossibleOutcomes(final Map<Answer, Map<Answer, Double>> newOutcomes) {
        this.possibleOutcomes = Optional.of(newOutcomes);
        return this;
    }

    /**
     * Builds the player.
     *
     * @return The HT player.
     * @throws GameException if creating the player is not allowed by the rules of the game.
     */
    @Override
    public IHTPlayer build() throws GameException {
        return new HTPlayer(
            this.name.orElseThrow(), 
            checkPossibleOutcomes(this.possibleOutcomes.orElseThrow()));
    }

    /**
     * Checks if all possible outcomes are defined for a player.
     *
     * @param outcomes The possible outcomes for the player.
     */
    private Map<Answer, Map<Answer, Double>> checkPossibleOutcomes(
            final Map<Answer, Map<Answer, Double>> outcomes) {
        this.checkPossibleOutcome(outcomes, Answer.HEADS, Answer.HEADS);
        this.checkPossibleOutcome(outcomes, Answer.HEADS, Answer.TAILS);
        this.checkPossibleOutcome(outcomes, Answer.HEADS, Answer.EDGE);
        this.checkPossibleOutcome(outcomes, Answer.TAILS, Answer.HEADS);
        this.checkPossibleOutcome(outcomes, Answer.TAILS, Answer.TAILS);
        this.checkPossibleOutcome(outcomes, Answer.TAILS, Answer.EDGE);
        this.checkPossibleOutcome(outcomes, Answer.EDGE, Answer.HEADS);
        this.checkPossibleOutcome(outcomes, Answer.EDGE, Answer.TAILS);
        this.checkPossibleOutcome(outcomes, Answer.EDGE, Answer.EDGE);
        return outcomes;
    }

    /**
     * Checks if a given outcome is defined for a player.
     *
     * @param outcomes     The possible outcomes for the player.
     * @param firstChoice  The choice of the first player.
     * @param secondChoice The choice of the second player.
     */
    private void checkPossibleOutcome(final Map<Answer, Map<Answer, Double>> outcomes, final Answer firstChoice,
            final Answer secondChoice) {
        if (outcomes.getOrDefault(firstChoice, Collections.emptyMap()).get(secondChoice) == null) {
            throw new IllegalArgumentException(
                String.format(
                    "No outcome defined for player '%s' and combination %s/%s.",
                    name,
                    firstChoice,
                    secondChoice));
        }
    }
}