package de.fhdw.gaming.ipspiel23.ht.domain.impl;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

import de.fhdw.gaming.core.domain.PlayerState;
import de.fhdw.gaming.ipspiel23.ht.domain.Answer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTPlayer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTState;

/**
 * Represents the state of a Heads or Tails game.
 */
final class HTState implements IHTState {
    
    /**
     * Represents the first player.
     */
    private final IHTPlayer playerOne;

    /**
     * Represents the second player.
     */
    private final IHTPlayer playerTwo; 
    
    /**
     * A map which maps the names of the players to the players.
     */
    private Map<String, IHTPlayer> playerMap;
    
    /**
     * A list of the players in the order in which they are added to the game.
     */
    private final List<IHTPlayer> orderedPlayers;
    
    /**
     * Creates a new {@link HTState} which is a deep copy of the given {@link HTState}.
     *
     * @param state The {@link HTState} to copy.
     */
    HTState(final HTState state) {
        this.playerOne = state.getPlayerOne().deepCopy();
        this.playerTwo = state.getPlayerTwo().deepCopy();
        this.orderedPlayers = new ArrayList<>();
        this.orderedPlayers.add(playerOne);
        this.orderedPlayers.add(playerTwo);
    }

    /**
     * Creates a new {@link HTState} with the given players.
     *
     * @param player1 The first player.
     * @param player2 The second player.
     * @throws IllegalArgumentException if the names of the players are identical.
     */
    public HTState(final IHTPlayer player1, final IHTPlayer player2) {
        this.playerOne = Objects.requireNonNull(player1, "player1");
        this.playerTwo = Objects.requireNonNull(player2, "player2");

        if (this.playerOne.getName().equalsIgnoreCase(this.playerTwo.getName())) {
            throw new IllegalArgumentException("Player names cannot be identical :P");
        }
        this.orderedPlayers = new ArrayList<>();
        this.orderedPlayers.add(this.playerOne);
        this.orderedPlayers.add(this.playerTwo);
    }

    /**
     * Returns the first player.
     */
    public IHTPlayer getPlayerOne() {
        return this.playerOne;
    }

    /**
     * Returns the second player.
     */
    public IHTPlayer getPlayerTwo() {
        return this.playerTwo;
    }
    
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof HTState) {
            final HTState other = (HTState) obj;
            return this.playerOne.equals(other.playerOne) && this.playerTwo.equals(other.playerTwo);
        }
        return false;
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.playerOne, this.playerTwo);
    }

    @Override
    public Map<String, IHTPlayer> getPlayers() {
        if (this.playerMap == null) {
            this.playerMap = new LinkedHashMap<>();
            this.playerMap.put(this.playerOne.getName(), this.playerOne);
            this.playerMap.put(this.playerTwo.getName(), this.playerTwo);
        }
        return playerMap;
    }

    @Override
    public Set<IHTPlayer> computeNextPlayers() {
        return getPlayers().values().stream()
                .filter(player -> player.getAnswer().isEmpty())
                .collect(Collectors.toSet());
    }

    @Override
    public void nextTurn() {
        if (computeNextPlayers().isEmpty()) {
            final List<Answer> answers = this.orderedPlayers.stream()
                .map(player -> player.getAnswer().orElseThrow())
                .collect(Collectors.toList());
            this.orderedPlayers.stream()
                .forEach(player -> applyOutcome(player, answers));
        }
    }

    /**
     * Applies the outcome of the game to the given player.
     *
     * @param player The player.
     * @param answers The answers of the players.
     */
    private static void applyOutcome(final IHTPlayer player, final List<Answer> answers) {
        final Double outcome = player.getPossibleOutcomes().get(answers.get(0)).get(answers.get(1));
        player.setState(outcomeToState(outcome));
        player.setOutcome(outcome);
    }

    @Override
    public IHTState deepCopy() {
        return new HTState(this);
    }

    /**
     * Converts an outcome to a {@link PlayerState}. 
     * @param outcome the 64-bit double precision floating point number to convert.
     * @return the new player state (DRAW if outcome == 0.0, WON if outcome > 0.0, LOST otherwise)
     */
    static PlayerState outcomeToState(final Double outcome) {
        return outcome > 0.0 
            ? PlayerState.WON 
            : outcome < 0.0 
                ? PlayerState.LOST 
                : PlayerState.DRAW;
    }

    @Override
    public IHTPlayer getFirstPlayer() {
        return this.playerOne;
    }

    @Override
    public IHTPlayer getSecondPlayer() {
        return this.playerTwo;
    }

    @Override
    public String toString() {
        return String.format("%s[firstPlayer=%s, secondPlayer=%s]", this.getClass().getSimpleName(), 
            this.playerOne, this.playerTwo);
    }
}